/*
	GnuJSP Toolkit - useful classes for JSP and servlet development
	Copyright (C) 1998, Vincent Partington <vincentp@xs4all.nl>

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

package	nl.nmg.sql;

import	java.sql.*;
import	java.util.Enumeration;
import	java.util.Properties;

/**
 * A proxy connection object encapsulates a real database connection.
 * Created by the PoolDriverManager.
 *
 * @see nl.nmg.sql.PoolDriverManager
 */
class PoolConnection implements Connection
{
	// real connection properties
	private Connection		realConn;
	private String			realURL;
	private Properties		realInfo;

	// is zero when connection is in use
	private long			lastUse;

	// allows real connection to be restored to original condition as best as possible
	private boolean			shouldRestoreAutoCommit = false,
							shouldRestoreCatalog = false,
							shouldRestoreReadOnly = false,
							shouldRestoreTransactionIsolation = false;
	private boolean			savedAutoCommit;
	private String			savedCatalog;
	private boolean			savedReadOnly;
	private int				savedTransactionIsolation;

	/**
	 * Create a new proxy connection.
	 *
	 * @param url	the real JDBC url
	 * @param info	the real JDBC connection properties
	 */
	PoolConnection(String url, Properties info) throws SQLException {
		realInfo = (Properties) info.clone();
		realConn = DriverManager.getConnection(url, info);
		realURL = url;
		lastUse = 0;
	}

	/**
	 * Let go of the proxy connection.
	 *
	 * <P><B>N.B.:</B> If a proxy connection object is kept
	 * after calling close(), its operation is unspecified.
	 */
	public void close() throws SQLException {
		// rollback any uncommited changes
		if(!realConn.getAutoCommit())
			realConn.rollback();

		// restore connection
		if(shouldRestoreAutoCommit) {
			shouldRestoreAutoCommit = false;
			realConn.setAutoCommit(savedAutoCommit);
		}
		if(shouldRestoreCatalog) {
			shouldRestoreCatalog = false;
			realConn.setCatalog(savedCatalog);
		}
		if(shouldRestoreReadOnly) {
			shouldRestoreReadOnly = false;
			realConn.setReadOnly(savedReadOnly);
		}
		if(shouldRestoreTransactionIsolation) {
			shouldRestoreTransactionIsolation = false;
			realConn.setTransactionIsolation(savedTransactionIsolation);
		}
		realConn.clearWarnings();

		lastUse = System.currentTimeMillis();
	}

	/**
	 * Try to grab this proxy connection. If succesfull, the connection
	 * is restored to its original state as best as possible
	 *
	 * @param url	the real JDBC url
	 * @param info	the real JDBC connection properties
	 */
	boolean use(String url, Properties info) throws SQLException {
		Enumeration		pnames;
		String			pname, pval;

		// is this proxy connection available and the correct one?
		if(lastUse == 0 || !url.equals(realURL) || isClosed())
			return false;

		// check if property hashes are identical
		pnames = realInfo.propertyNames();
		while(pnames.hasMoreElements()) {
			pname = (String) pnames.nextElement();
			pval = realInfo.getProperty(pname);
			if(pval == null || !pval.equals(info.getProperty(pname))) {
				return false;
			}
		}

		pnames = info.propertyNames();
		while(pnames.hasMoreElements()) {
			pname = (String) pnames.nextElement();
			pval = info.getProperty(pname);
			if(pval == null || !pval.equals(realInfo.getProperty(pname))) {
				return false;
			}
		}

		// is this proxy connection _really_ available?
		synchronized(this) {
			if(lastUse == 0) {
				return false;
			}
			lastUse = 0;
		}

		return true;
	}

	/**
	 * Close real connection and destroy proxy connection
	 * if proxy connection has not been used in a while.
	 *
	 * @param inactiveSince		XXX enter correct description XXX
	 */
	boolean destroyIf(long inactiveSince) {
		if(lastUse != 0 && lastUse < inactiveSince) {
			synchronized(this) {
				if(lastUse != 0 && lastUse < inactiveSince) {
					try {
						realConn.close();
					} catch(SQLException exc) { }
					realConn = null;
					return true;
				}
			}
		}

		return false;
	}

	/**
	 * Called when object is about to be garbage collected.
	 */
	protected void finalize() {
		try {
			realConn.close();
		} catch(SQLException exc) { }
	}

	/*
	 * Proxy methods; the corresponding method in the
	 * real connection is called for each of these methods.
	 */

	public void clearWarnings() throws SQLException {
		realConn.clearWarnings();
	}

	public void commit() throws SQLException {
		realConn.commit();
	}

	public Statement createStatement() throws SQLException {
		return realConn.createStatement();
	}

	public boolean getAutoCommit() throws SQLException {
		return realConn.getAutoCommit();
	}

	public String getCatalog() throws SQLException {
		return realConn.getCatalog();
	}

	public DatabaseMetaData getMetaData() throws SQLException {
		return realConn.getMetaData();
	}

	public int getTransactionIsolation() throws SQLException {
		return realConn.getTransactionIsolation();
	}

	public SQLWarning getWarnings() throws SQLException {
		return realConn.getWarnings();
	}

	public boolean isClosed() throws SQLException {
		return realConn == null || realConn.isClosed();
	}

	public boolean isReadOnly() throws SQLException {
		return realConn.isReadOnly();
	}

	public String nativeSQL(String sql) throws SQLException {
		return realConn.nativeSQL(sql);
	}

	public CallableStatement prepareCall(String sql) throws SQLException {
		return realConn.prepareCall(sql);
	}

	public PreparedStatement prepareStatement(String sql) throws SQLException {
		return realConn.prepareStatement(sql);
	}

	public void rollback() throws SQLException {
		realConn.rollback();
	}

	public void setAutoCommit(boolean autoCommit) throws SQLException {
		synchronized(this) {
			if(!shouldRestoreAutoCommit) {
				shouldRestoreAutoCommit = true;
				savedAutoCommit = realConn.getAutoCommit();
			}
		}

		realConn.setAutoCommit(autoCommit);
	}

	public void setCatalog(String catalog) throws SQLException {
		synchronized(this) {
			if(!shouldRestoreCatalog) {
				shouldRestoreCatalog = true;
				savedCatalog = realConn.getCatalog();
			}
		}

		realConn.setCatalog(catalog);
	}

	public void setReadOnly(boolean readOnly) throws SQLException {
		synchronized(this) {
			if(!shouldRestoreReadOnly) {
				shouldRestoreReadOnly = true;
				savedReadOnly = realConn.isReadOnly();
			}
		}

		realConn.setReadOnly(readOnly);
	}

	public void setTransactionIsolation(int level) throws SQLException {
		synchronized(this) {
			if(!shouldRestoreTransactionIsolation) {
				shouldRestoreTransactionIsolation = true;
				savedTransactionIsolation = realConn.getTransactionIsolation();
			}
		}

		realConn.setTransactionIsolation(level);
	}
}
