/*
	GnuJSP Toolkit - useful classes for JSP and servlet development
	Copyright (C) 1998, Vincent Partington <vincentp@xs4all.nl>

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

package	nl.nmg.sql;

import	java.sql.*;
import	java.util.Enumeration;
import	java.util.Properties;
import	java.util.Vector;

/**
 * <P>The PoolDriverManager keeps a pool of connections. If a
 * connection is requested, the pool is checked for a similar
 * free connection. If none can be found, a new connection is made.
 *
 * <P>The DriverManager.getConnection() methods have been
 * reimplemented by PoolDriverManager, allowing you to use
 * it like this:
 *
 * <PRE>
 * conn = PoolDriverManager.getConnection("jdbc:odbc:SampleTable");
 * </PRE>
 *
 * <P>A connection from the pool is considered in use, until its
 * close() method is called. This does not actually close the
 * connection, but returns the connection to the pool instead.
 * (The reference used to call close() should not be used after
 * that; it will refer to the connection in the pool and use of
 * this connection will result in unpredictable behaviour.)
 *
 * <P>When a connection is returned to the pool, any uncommited
 * changes are rolled back. Persistent changes made to the
 * connection through the Connection interface are undone,
 * but persistent changes made through SQL commands are not.
 *
 * <P>If a connection in the pool has not been used for 5 minutes
 * it is closed and removed from the pool. A "pool-cleaner" thread
 * checks the state of the connections in the pool every 30 seconds.
 *
 * @see java.sql.DriverManager
 * @see nl.nmg.sql.PoolConnection
 * @see nl.nmg.sql.PoolDriver
 */
public class PoolDriverManager implements Runnable {
	static int				MAJOR_VERSION = 1,
							MINOR_VERSION = 0;
	private static int		POOLCLEANER_MILLIS = 30000,		// 30 seconds
							TIMEOUT_MILLIS = 5 * 60000;		// 5 minutes

	/*
	 * Initialize pool and cleaner thread
	 */
	private static Vector	poolConnections;
	private static Thread	poolCleanerThread;

	static
	{
		poolConnections = new Vector();
		poolCleanerThread = new Thread(new PoolDriverManager(), "poolcleaner");
		poolCleanerThread.setPriority(Thread.MIN_PRIORITY);
		poolCleanerThread.setDaemon(true);
		poolCleanerThread.start();
	}

	/**
	 * No public constructor
	 */
	private PoolDriverManager()
	{
		;
	}

	/**
	 * Attempt to establish a connection to the given database URL.
	 * The PoolDriverManager attempts to get a connection from the
	 * pool and calls DriverManager.getConnection() if that fails.
	 *
	 * @param url a database url of the form jdbc:<em>subprotocol</em>:<em>subname</em>
	 * @param info a list of arbitrary string tag/value pairs as
	 * connection arguments; normally at least a "user" and
	 * "password" property should be included
	 * @return a Connection to the URL 
	 * @exception SQLException if a database-access error occurs.
	 */
	public static Connection getConnection(String url, Properties info) throws SQLException
	{
		Enumeration		pcs;
		PoolConnection	pc;

		pcs = poolConnections.elements();
		while(pcs.hasMoreElements()) {
			pc = (PoolConnection) pcs.nextElement();
			if(pc == null) { // why this test?
				break;
			}

			if(pc.use(url, info)) {
				return pc;
			}
		}

		pc = new PoolConnection(url, info);
		poolConnections.addElement(pc);
// FIXME: how expensive is resume() when it is not necessary?
		poolCleanerThread.resume();
		return pc;
	}

	/**
	 * Attempt to establish a connection to the given database URL.
	 * The PoolDriverManager attempts to get a connection from the
	 * pool and calls DriverManager.getConnection() if that fails.
	 *
	 * @param url a database url of the form jdbc:<em>subprotocol</em>:<em>subname</em>
	 * @param user the database user on whose behalf the Connection is being made
	 * @param password the user's password
	 * @return a Connection to the URL 
	 * @exception SQLException if a database-access error occurs.
	 */
	public static Connection getConnection(String url, String user, String password) throws SQLException
	{
		Properties		info = new Properties();

		info.put("user", user);
		info.put("password", password);
		return getConnection(url, info);
	}

	/**
	 * Attempt to establish a connection to the given database URL.
	 * The PoolDriverManager attempts to get a connection from the
	 * pool and calls DriverManager.getConnection() if that fails.
	 *
	 * @param url a database url of the form jdbc:<em>subprotocol</em>:<em>subname</em>
	 * @return a Connection to the URL 
	 * @exception SQLException if a database-access error occurs.
	 */
	public static Connection getConnection(String url) throws SQLException
	{
		return getConnection(url, new Properties());
	}

	/**
	 * The pool cleaner thread. Checks for connections that have not been used
	 * for more than 5 minutes, every 30 seconds.
	 */
	public void run()
	{
		long				inactiveSince;
		Enumeration			pcs;
		PoolConnection		pc;

		ROUND: for(;;) {
			// suspend if there are no connection to monitor
			if(poolConnections.size() == 0) {
				poolCleanerThread.suspend();
			}

			// check for inactive connections
			pcs = poolConnections.elements();
			inactiveSince = System.currentTimeMillis() - TIMEOUT_MILLIS;
			while(pcs.hasMoreElements()) {
				pc = (PoolConnection) pcs.nextElement();
				if(pc.destroyIf(inactiveSince)) {
					poolConnections.removeElement(pc);
					continue ROUND;
				}
			}

			// wait for next poolcleaning round
			try {
				Thread.sleep(POOLCLEANER_MILLIS);
			} catch(InterruptedException exc) { }
		}
	}
}
